
/* Continuation objects */

#include "Python.h"
#include "continue.h"

extern CONTINUATION * root_continuation;

static PyObject *ErrorObject;

typedef struct {
  PyObject_HEAD
  CONTINUATION	*cont;
  CONTINUATION	*root;
} ContinuationObject;

staticforward PyTypeObject Continuation_Type;

#define ContinuationObject_Check(v)	((v)->ob_type == &Continuation_Type)

static ContinuationObject *
newContinuationObject(arg)
     PyObject *arg;
{
  ContinuationObject *self;
  self = PyObject_NEW(ContinuationObject, &Continuation_Type);
  if (self == NULL)
    return NULL;
  return self;
}

/* Continuation methods */

static void
Continuation_dealloc(self)
     ContinuationObject *self;
{
  free_continuation (self->cont);
  PyMem_DEL(self);
}

static PyObject *
Continuation_demo(self, args)
     ContinuationObject *self;
     PyObject *args;
{
  if (!PyArg_ParseTuple(args, ""))
    return NULL;
  Py_INCREF(Py_None);
  return Py_None;
}

static PyObject *
Continuation_throw (self, args)
     ContinuationObject *self;
     PyObject *args;
{
  PyObject * val;
  if (!PyArg_ParseTuple (args, "O", &val)) {
    return NULL;
  } else {
    throw_to_continuation (self->cont, val, self->root);
  }
}

static PyMethodDef Continuation_methods[] = {
  {"throw",	(PyCFunction)Continuation_throw,	1},
  {NULL,		NULL}		/* sentinel */
};

static PyObject *
Continuation_getattr(self, name)
     ContinuationObject *self;
     char *name;
{
  return Py_FindMethod(Continuation_methods, (PyObject *)self, name);
}

staticforward PyTypeObject Continuation_Type = {
	PyObject_HEAD_INIT(&PyType_Type)
	0,			/*ob_size*/
	"Continuation",			/*tp_name*/
	sizeof(ContinuationObject),	/*tp_basicsize*/
	0,			/*tp_itemsize*/
	/* methods */
	(destructor)Continuation_dealloc, /*tp_dealloc*/
	0,			/*tp_print*/
	(getattrfunc)Continuation_getattr, /*tp_getattr*/
	0,			/*tp_setattr*/
	0,			/*tp_compare*/
	0,			/*tp_repr*/
	0,			/*tp_as_number*/
	0,			/*tp_as_sequence*/
	0,			/*tp_as_mapping*/
	0,			/*tp_hash*/
};
/* --------------------------------------------------------------------- */

#if 0
static PyObject *
callcc_make_root(self, args)
     PyObject *self; /* Not used */
     PyObject *args;
{
  ContinuationObject *k_root;
  if (!PyArg_ParseTuple(args, "")) {
    return NULL;
  } else { 
    k_root = newContinuationObject(args);
    if ( k_root == NULL ) {
      return NULL;
    } else {
      CONTINUATION * k = make_root_continuation (&k_root);
      if (!k) {
	Py_DECREF (k_root);
	return NULL;
      } else {
	k_root->cont = k;
	k_root->root = NULL;
	setjmp (k->jmpbuf);
	return (PyObject *) k_root;
      }
    }
  }
}
#endif

static PyObject *
callcc_new(self, args)
     PyObject *self; /* Not used */
     PyObject *args;
{
  ContinuationObject *cc;
	
  if (!PyArg_ParseTuple(args, "")) {
    return NULL;
  } else { 
    cc = newContinuationObject(args);
    if ( cc == NULL ) {
      return NULL;
    } else {
      CONTINUATION * k = make_continuation (root_continuation);
      if (!k) {
	Py_DECREF (cc);
	return NULL;
      } else {
	cc->cont = k;
	cc->root = root_continuation;
	setjmp (k->jmpbuf);
	return (PyObject *) cc;
      }
    }
  }
}

/* List of functions defined in the module */

static PyMethodDef callcc_methods[] = {
  /*  {"make_root",		callcc_make_root,	1}, */
  {"new",		callcc_new,		1},
  {NULL,		NULL}		/* sentinel */
};


/* Initialization function for the module (*must* be called initcallcc) */

void
initcallcc()
{
  PyObject *m, *d;

  /* Create the module and add the functions */
  m = Py_InitModule("callcc", callcc_methods);

  /* Add some symbolic constants to the module */
  d = PyModule_GetDict(m);
  ErrorObject = PyErr_NewException("callcc.error", NULL, NULL);
  PyDict_SetItemString(d, "error", ErrorObject);
}
